@@ -7,11 +7,11 @@ from cacheout import CacheManager
 from msal import PublicClientApplication
 from prometheus_client import Counter
 
-from openvpn_auth_azure_ad import util
-from openvpn_auth_azure_ad._version import __version__
-from openvpn_auth_azure_ad.openvpn import OpenVPNManagementInterface
-from openvpn_auth_azure_ad.util import errors
-from openvpn_auth_azure_ad.util.thread_pool import ThreadPoolExecutorStackTraced
+from . import util
+from ._version import __version__
+from .openvpn import OpenVPNManagementInterface
+from .util import errors
+from .util.thread_pool import ThreadPoolExecutorStackTraced
 
 openvpn_auth_azure_ad_events = Counter(
     "openvpn_auth_azure_ad_events", "track events", ["event"]
@@ -43,19 +43,25 @@ class AADAuthenticator(object):
         app: PublicClientApplication,
         graph_endpoint: str,
         authenticators: str,
-        verify_common_name: bool,
+        openvpn_identity_key: str,
+        verify_openvpn_client: bool,
+        verify_openvpn_client_id_token_claim: bool,
         auth_token: bool,
         auth_token_lifetime: int,
+        remember_user: bool,
         threads: int,
         host: str = None,
         port: int = None,
         socket: str = None,
         password: str = None,
+        release_hold: bool = None,
     ):
         self._app = app
         self._graph_endpoint = graph_endpoint
         self._authenticators = [s.strip() for s in authenticators.split(",")]
-        self._openvpn = OpenVPNManagementInterface(host, port, socket, password)
+        self._openvpn = OpenVPNManagementInterface(
+            release_hold, host, port, socket, password
+        )
         self._openvpn.connect()
         self._states = CacheManager(
             {
@@ -65,9 +71,14 @@ class AADAuthenticator(object):
             }
         )
 
-        self._verify_common_name_enabled = verify_common_name
+        self._openvpn_identity_key = openvpn_identity_key
+        self._verify_openvpn_client = verify_openvpn_client
+        self._verify_openvpn_client_id_token_claim = (
+            verify_openvpn_client_id_token_claim
+        )
         self._auth_token_enabled = auth_token
         self._auth_token_lifetime = auth_token_lifetime
+        self._remember_user_enabled = remember_user
         self._thread_pool = ThreadPoolExecutorStackTraced(max_workers=threads)
 
     def run(self) -> None:
@@ -77,6 +88,7 @@ class AADAuthenticator(object):
                 message = self._openvpn.receive()
                 if not message:
                     logger.error("Connection to OpenVPN closed. Reconnecting...")
+                    self._openvpn.disconnect(False)
                     self._openvpn.connect(True)
                     continue
 
@@ -102,9 +114,9 @@ class AADAuthenticator(object):
 
     def send_authentication_success(self, client: dict) -> None:
         self.log_info(client, "authentication succeeded")
-        if client["auth_token"] is not None:
+        if self._auth_token_enabled:
             self._openvpn.send_command(
-                'client-auth %s %s\npush "auth-token %s"\nEND'
+                "\n".join(("client-auth %s %s", 'push "auth-token %s"', "END"))
                 % (client["cid"], client["kid"], client["auth_token"])
             )
         else:
@@ -135,6 +147,11 @@ class AADAuthenticator(object):
                 % (client["cid"], client["kid"], message, client_challenge)
             )
 
+    def send_authentication_pending(self, client: dict, timeout: int) -> None:
+        self._openvpn.send_command(
+            "client-pending-auth %s %s %d" % (client["cid"], client["kid"], timeout)
+        )
+
     def setup_auth_token(self, client) -> None:
         if self._auth_token_enabled:
             client["auth_token"] = util.generated_id()
@@ -142,19 +159,52 @@ class AADAuthenticator(object):
                 client["auth_token"], client["state_id"], self._auth_token_lifetime
             )
 
-    def verify_common_name(self, client, result) -> bool:
-        if (
-            "common_name" not in client["env"]
-            or "id_token_claims" not in result
-            or "preferred_username" not in result["id_token_claims"]
-        ):
+    def verify_openvpn_client(self, client, result) -> bool:
+        if not self._verify_openvpn_client:
+            return True
+
+        if self._openvpn_identity_key not in client["env"]:
+            self.log_warn(
+                client, "Missing %s in client env" % (self._openvpn_identity_key,)
+            )
             return False
 
-        return (
-            client["env"]["common_name"]
-            == result["id_token_claims"]["preferred_username"]
+        if "id_token_claims" not in result:
+            self.log_warn(client, "Could not get id_token_claims")
+            return False
+
+        if self._verify_openvpn_client_id_token_claim not in result["id_token_claims"]:
+            self.log_warn(
+                client,
+                "claim %s does not exist in id_token"
+                % (self._verify_openvpn_client_id_token_claim,),
+            )
+            return False
+
+        verify_result = (
+            client["env"][self._openvpn_identity_key]
+            == result["id_token_claims"][self._verify_openvpn_client_id_token_claim]
         )
 
+        if not verify_result:
+            self.log_info(
+                client,
+                (
+                    "OpenVPN %s does not match Azure AD user (claim: %s). "
+                    + "Client certificate: '%s' - Azure AD: '%s'"
+                )
+                % (
+                    self._openvpn_identity_key,
+                    self._verify_openvpn_client_id_token_claim,
+                    client["env"][self._openvpn_identity_key],
+                    result["id_token_claims"][
+                        self._verify_openvpn_client_id_token_claim
+                    ],
+                ),
+            )
+
+        return verify_result
+
     def handle_reauth(self, client: Dict) -> Optional[dict]:
         if not self._auth_token_enabled:
             return None
@@ -185,8 +235,8 @@ class AADAuthenticator(object):
         # recheck if connected client is the same as inside the state.
         if (
             client["auth_token"] != authenticated_state["client"]["auth_token"]
-            or client["env"]["common_name"]
-            != authenticated_state["client"]["env"]["common_name"]
+            or client["env"][self._openvpn_identity_key]
+            != authenticated_state["client"]["env"][self._openvpn_identity_key]
             or client["env"]["username"]
             != authenticated_state["client"]["env"]["username"]
         ):
@@ -203,7 +253,7 @@ class AADAuthenticator(object):
         if not client["env"]["password"].startswith("CRV1::"):
             return None
 
-        # get state id from client
+        # get state id from the client
         state_id = util.get_state_id(client)
         if state_id is None:
             return None
@@ -213,6 +263,8 @@ class AADAuthenticator(object):
         if state is None:
             return None
 
+        client["state_id"] = state_id
+
         # delete used challenge state
         self._states["challenge"].delete(state_id)
 
@@ -244,21 +296,14 @@ class AADAuthenticator(object):
 
     def authenticate_client(self, client: dict) -> None:
         result = {}
-
         if client["reason"] == "reauth":
-            # allow clients to bypass azure ad authentication by reauthenticate via auth-token
+            # allow clients to bypass azure ad authentication by reauthenticate via an auth-token
             result = self.handle_reauth(client)
             if result is not None:
                 if util.is_authenticated(result):
-                    if (
-                        self._verify_common_name_enabled
-                        and not self.verify_common_name(client, result)
-                    ):
-                        self.log_info(
-                            client, "common_name does not match Azure AD username."
-                        )
+                    if not self.verify_openvpn_client(client, result):
                         self.send_authentication_error(
-                            client, "common_name_not_matched", None
+                            client, "client_certificate_not_matched", None
                         )
                         return None
 
@@ -274,15 +319,9 @@ class AADAuthenticator(object):
             if result is not None:
                 client["state_id"] = util.get_state_id(client)
                 if util.is_authenticated(result):
-                    if (
-                        self._verify_common_name_enabled
-                        and not self.verify_common_name(client, result)
-                    ):
-                        self.log_info(
-                            client, "common_name does not match Azure AD username."
-                        )
+                    if not self.verify_openvpn_client(client, result):
                         self.send_authentication_error(
-                            client, "common_name_not_matched", None
+                            client, "client_certificate_not_matched", None
                         )
                         return None
 
@@ -300,6 +339,7 @@ class AADAuthenticator(object):
                     openvpn_auth_azure_ad_auth_failures.labels(
                         AADAuthenticatorFlows.DEVICE_TOKEN
                     ).inc()
+
                     if 70016 in result.get("error_codes", []):
                         error = "device token flow errored: no user action"
                         self.log_info(client, error)
@@ -316,6 +356,24 @@ class AADAuthenticator(object):
                     return
 
         client["state_id"] = util.generated_id()
+        if self._remember_user_enabled:
+            if self._openvpn_identity_key in client["env"]:
+                accounts = self._app.get_accounts(
+                    client["env"][self._openvpn_identity_key]
+                )
+                if accounts:
+                    result = self._app.acquire_token_silent(
+                        self.token_scopes, account=accounts[0]
+                    )
+
+            if util.is_authenticated(result):
+                self.setup_auth_token(client)
+                self._states["authenticated"].set(
+                    client["state_id"], {"client": client, "result": result}
+                )
+                self.send_authentication_success(client)
+                return
+
         if AADAuthenticatorFlows.USER_PASSWORD in self._authenticators:
             self.log_debug(client, "Authenticate using username/password flow")
             openvpn_auth_azure_ad_auth_total.labels(
@@ -328,25 +386,23 @@ class AADAuthenticator(object):
             )
 
             if util.is_authenticated(result):
-                if self._verify_common_name_enabled and not self.verify_common_name(
-                    client, result
-                ):
-                    self.log_info(
-                        client, "common_name does not match Azure AD username."
-                    )
+                if not self.verify_openvpn_client(client, result):
                     self.send_authentication_error(
-                        client, "common_name_not_matched", None
+                        client, "client_certificate_not_matched", None
                     )
                     return None
             else:
                 openvpn_auth_azure_ad_auth_failures.labels(
                     AADAuthenticatorFlows.USER_PASSWORD
                 ).inc()
+
                 if 65001 in result.get("error_codes", []):
                     # AAD requires user consent for U/P flow
                     error = (
                         "Get consent first: %s"
-                        % self._app.get_authorization_request_url(self.token_scopes)
+                        % self._app.initiate_auth_code_flow(self.token_scopes)[
+                            "auth_uri"
+                        ]
                     )
                     self.send_authentication_error(client, "consent_needed", error)
                 else:
@@ -360,14 +416,13 @@ class AADAuthenticator(object):
                 AADAuthenticatorFlows.USER_PASSWORD
             ).inc()
             self.log_info(client, "password flow succeeded")
-            # Do not send successful auth command, because device token can be enabled.
+            # Do not send a successful auth command, because a device token can be enabled.
 
         if AADAuthenticatorFlows.DEVICE_TOKEN in self._authenticators:
             self.log_info(client, "Start to authenticate using device token flow")
             flow = self.device_auth_start()
-            message = flow["message"] + " Then press OK here. No input required here."
+            message = flow["message"] + " Then press OK here."
 
-            self.log_debug(client, "Save state as %s" % client["state_id"])
             self._states["challenge"].set(client["state_id"], {"flow": flow})
 
             openvpn_auth_azure_ad_auth_total.labels(
@@ -394,8 +449,8 @@ class AADAuthenticator(object):
         return flow
 
     def device_token_finish(self, state: dict) -> dict:
-        # Block authentication at least for 30 seconds
-        state["flow"]["expires_in"] = 30
+        # Block authentication at least for 120 seconds
+        state["flow"]["expires_in"] = 60
         state["flow"]["expires_at"] = time.time() + state["flow"]["expires_in"]
 
         return self._app.acquire_token_by_device_flow(state["flow"])
@@ -435,19 +490,27 @@ class AADAuthenticator(object):
 
         return client
 
-    def log_info(self, client: dict, message: str) -> None:
-        prefix = self.format_log_prefix(client)
+    def format_log_prefix(self, client: dict) -> str:
+        prefix = "cid: %s" % (client["cid"],)
+        if self._openvpn_identity_key in client["env"]:
+            prefix += " | %s" % client["env"][self._openvpn_identity_key]
 
-        logger.info("[%s]: %s" % (prefix, message))
+        if "state_id" in client and client["state_id"] is not None:
+            prefix += " | %s" % client["state_id"]
+
+        return prefix
 
     def log_debug(self, client: dict, message: str) -> None:
         prefix = self.format_log_prefix(client)
 
         logger.debug("[%s]: %s" % (prefix, message))
 
-    def format_log_prefix(self, client: dict) -> str:
-        prefix = "cid: %s" % (client["cid"],)
-        if "common_name" in client["env"]:
-            prefix += " | %s" % client["env"]["common_name"]
+    def log_info(self, client: dict, message: str) -> None:
+        prefix = self.format_log_prefix(client)
 
-        return prefix
+        logger.info("[%s]: %s" % (prefix, message))
+
+    def log_warn(self, client: dict, message: str) -> None:
+        prefix = self.format_log_prefix(client)
+
+        logger.warning("[%s]: %s" % (prefix, message))
